use super::{
    VerbContext,
    UserVerb,
    UserVerbRef,
    UResult,
    user_error,
    get_user_or_fail,
    get_player_item_or_fail,
    search_item_for_user,
    parsing::parse_command_name,
};
use crate::{
    models::corp::{CorpMembership, CorpPermission},
    db::ItemSearchParams,
};
use chrono::Utc;
use async_trait::async_trait;
use ansi::ansi;

fn check_corp_perm(perm: &CorpPermission, mem: &CorpMembership) -> bool {
    mem.permissions.iter().any(|p| *p == CorpPermission::Holder || *p == *perm)
}

async fn corp_invite(ctx: &mut VerbContext<'_>, remaining: &str) -> UResult<()> {
    let (target_raw, into_raw) = match remaining.rsplit_once(" into ") {
        None => user_error(
            ansi!("Usage: <bold>corp hire<reset> username <bold>into<reset> corpname").to_owned()
        )?,
        Some(c) => c
    };
    let user = get_user_or_fail(ctx)?;
    let player = get_player_item_or_fail(ctx).await?;
    let (corp_id, corp, mem) =
        match ctx.trans.match_user_corp_by_name(into_raw.trim(), &user.username).await? {
            None => user_error("No such corp!".to_owned())?,
            Some(c) => c
        };
    if !check_corp_perm(&CorpPermission::Hire, &mem) || mem.joined_at.is_none() {
        user_error("You don't have hire permissions for that corp".to_owned())?;
    }

    let target_user = search_item_for_user(ctx, &ItemSearchParams {
        include_loc_contents: true,
        ..ItemSearchParams::base(&player, target_raw.trim())
    }).await?;

    if target_user.item_type != "player" {
        user_error("Only players can be hired.".to_owned())?;
    }
    
    let (their_sess, their_sess_dat) = match ctx.trans.find_session_for_player(&target_user.item_code).await? {
        None => user_error("The user needs to be logged in while you hire them.".to_owned())?,
        Some(c) => c
    };

    ctx.trans.expire_old_invites().await?;

    match ctx.trans.match_user_corp_by_name(into_raw.trim(), &target_user.item_code).await? {
        None => (),
        Some((_, _, CorpMembership { invited_at: Some(_), .. })) =>
            user_error("They've already been invited.".to_owned())?,
        Some((_, _, _)) =>
            user_error("They're already hired.".to_owned())?,
    };
    
    let new_mem = CorpMembership {
        invited_at: Some(Utc::now()),
        allow_combat: corp.allow_combat_required,
        ..Default::default()
    };
    ctx.trans.upsert_corp_membership(&corp_id, &target_user.item_code, &new_mem).await?;
    
    ctx.trans.queue_for_session(&their_sess, Some(&format!(
        ansi!("{} wants to hire you into {}! Type <bold>corp join {}<reset> to accept.{}\n"),
        &player.display_for_sentence(!their_sess_dat.less_explicit_mode, 1, true),
        &corp.name,
        &corp.name,
        if new_mem.allow_combat {
            " This corp is configured to allow the leadership to declare war on other corps; if you join, you may be attacked by members of other corps, even without your personal consent."
        } else {
            ""
        }
    ))).await?;
    ctx.trans.queue_for_session(
        &ctx.session,
        Some(&format!(
            "You offer to hire {} into {}.\n",
            &target_user.display_for_sentence(!their_sess_dat.less_explicit_mode, 1, false),
            &corp.name
        ))).await?;
    Ok(())
}

async fn corp_join(_ctx: &mut VerbContext<'_>, _remaining: &str) -> UResult<()> {
    user_error("Join not implemented yet".to_owned())?
}

async fn corp_list(ctx: &mut VerbContext<'_>, _remaining: &str) -> UResult<()> {
    let mut msg = String::new();
    let user = get_user_or_fail(ctx)?;

    let corps = ctx.trans.get_corp_memberships_for_user(&user.username).await?;
    if corps.is_empty() {
        msg.push_str(ansi!(
            "You don't yet belong to any corps - try <bold>who<reset> to see the \
             corps of people online, and message someone in a corp to see if they \
             will hire you! Or buy a new corp licence at the Kings Office in Melbs.\n"
        ));
    } else {
        msg.push_str("You belong to the following corps:\n"); 
        msg.push_str(&format!(ansi!("<bgblue><white><bold>| {:20} | {:7} | {:5} |<reset>\n"),
                              "Name", "Combat?", "Order"));
        for corp in &corps {
            match corp {
                (_, name, CorpMembership { allow_combat: combat, priority: p, .. }) => {
                    msg.push_str(&format!("| {:20} | {:7} | {:5} |\n",
                                          name, if *combat { "Y" } else { "N" }, p));
                }
            }
        }
   }
    
    ctx.trans.queue_for_session(ctx.session, Some(&msg)).await?;
    Ok(())
}

pub struct Verb;
#[async_trait]
impl UserVerb for Verb {
    async fn handle(self: &Self, ctx: &mut VerbContext, _verb: &str, remaining: &str) -> UResult<()> {
        let (command, remaining) = parse_command_name(remaining);
        match command.to_lowercase().as_str() {
            "hire" | "invite" => corp_invite(ctx, remaining).await?,
            "join" => corp_join(ctx, remaining).await?,
            "" | "list" => corp_list(ctx, remaining).await?,
            _ => user_error("Unknown command".to_owned())?
        }
        Ok(())
    }
}
static VERB_INT: Verb = Verb;
pub static VERB: UserVerbRef = &VERB_INT as UserVerbRef;
